#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
from string import Formatter
from typing import Any, Sequence, List, Dict, Union, Optional, Generator
from dataclasses import dataclass, KW_ONLY, asdict
from json import loads as json_loads
import warnings

from ibm_watson_machine_learning._wrappers import requests
from ibm_watson_machine_learning.helpers import DataConnection
from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.wml_client_error import WMLClientError, InvalidMultipleArguments, InvalidValue
from ibm_watson_machine_learning.utils import NextResourceGenerator
from ibm_watson_machine_learning.utils.autoai.utils import load_file_from_file_system_nonautoai
from ibm_watson_machine_learning.utils.autoai.enums import DataConnectionTypes
from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes
from ibm_watson_machine_learning.lifecycle import SpecStates
from ibm_watson_machine_learning import APIClient


@dataclass
class PromptTuningParams:
    base_model: dict
    _: KW_ONLY
    accumulate_steps: int = None
    batch_size: int = None
    init_method: str = None
    init_text: str = None
    learning_rate: float = None
    max_input_tokens: int = None
    max_output_tokens: int = None
    num_epochs: int = None
    task_id: str = None
    tuning_type: str = None
    verbalizer: str = None

    def to_dict(self):
        return {key: value for key, value in asdict(self).items() if value is not None}


def _get_foundation_models_spec(url, operation_name, additional_params: dict = None):
    params = {"version": "2023-09-30"}
    if additional_params:
        params.update(additional_params)
    response = requests.get(url,
                            params=params,
                            headers={'X-WML-User-Client': 'PythonClient'})
    if response.status_code == 200:
        return response.json()
    elif response.status_code == 404:
        raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_no_foundation_models"), logg_messages=False)
    else:
        msg = f"{operation_name} failed. Reason: {response.text}"
        raise WMLClientError(msg)


def get_model_specs(url: str, model_id: Optional[str] = None) -> dict:
    """
    Operations to query the details of the deployed foundation models.

    :param url: environment url
    :type url: str

    :param model_id: Id of the model, defaults to None (all models specs are returned).
    :type model_id: Optional[str, ModelTypes], optional

    :return: list of deployed foundation model specs
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import get_model_specs

        # GET ALL MODEL SPECS
        get_model_specs(
            url="https://us-south.ml.cloud.ibm.com"
            )

        # GET MODEL SPECS BY MODEL_ID
        get_model_specs(
            url="https://us-south.ml.cloud.ibm.com",
            model_id="google/flan-ul2"
            )
    """
    try:
        if model_id:
            if isinstance(model_id, ModelTypes):
                model_id = model_id.value

            try:
                return [res for res in _get_foundation_models_spec(f"{url}/ml/v1/foundation_model_specs",
                                                                   "Get available foundation models")['resources']
                        if res['model_id'] == model_id][0]
            except WMLClientError:  # Remove on CPD 5.0 release
                return [res for res in _get_foundation_models_spec(f"{url}/ml/v1-beta/foundation_model_specs",
                                                                   "Get available foundation models")['resources']
                        if res['model_id'] == model_id][0]
        else:
            try:
                return _get_foundation_models_spec(f"{url}/ml/v1/foundation_model_specs",
                                                   "Get available foundation models")
            except WMLClientError:  # Remove on CPD 5.0 release
                return _get_foundation_models_spec(f"{url}/ml/v1-beta/foundation_model_specs",
                                                   "Get available foundation models")
    except WMLClientError as e:
        raise WMLClientError(Messages.get_message(url, message_id="fm_prompt_tuning_no_model_specs"), e)


def get_custom_model_specs(credentials: dict = None,
                           api_client: 'APIClient' = None,
                           model_id: Optional[str] = None,
                           limit: int = 100, 
                           asynchronous: bool = False, 
                           get_all: bool = False,
                           verify=None) -> Union[dict, Generator[dict, None, None]]:
    """Get details on available custom model(s) as dict or as generator (``asynchronous``). 
    If ``asynchronous`` or ``get_all`` is set, then ``model_id`` is ignored.

    :param credentials: credentials to watsonx.ai instance
    :type credentials: dict, optional

    :param api_client: API client to connect to service
    :type api_client: APIClient, optional

    :param model_id: Id of the model, defaults to None (all models specs are returned).
    :type model_id: str, optional

    :param limit:  limit number of fetched records. Possible values: 1 ≤ value ≤ 200, default value: 100
    :type limit: int, optional

    :param asynchronous:  if True, it will work as a generator
    :type asynchronous: bool, optional

    :param get_all:  if True, it will get all entries in 'limited' chunks
    :type get_all: bool, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    :return: details of supported custom models
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import get_custom_model_specs

        get_custom_models_spec(api_client=client)
        get_custom_models_spec(credentials=credentials)
        get_custom_models_spec(api_client=client, model_id='mistralai/Mistral-7B-Instruct-v0.2')
        get_custom_models_spec(api_client=client, limit=20)
        get_custom_models_spec(api_client=client, limit=20, get_all=True)
        for spec in get_custom_model_specs(api_client=client, limit=20, asynchronous=True, get_all=True):
            print(spec, end="")

    """
    warnings.warn("Model needs to be first stored via client.repository.store_model(model_id, meta_props=metadata)"
                  " and deployed via client.deployments.create(asset_id, metadata) to be used.")

    if credentials is None and api_client is None:
        raise InvalidMultipleArguments(params_names_list=["credentials", "api_client"],
                                       reason="None of the arguments were provided.")
    elif credentials:
        client = APIClient(wml_credentials=credentials, verify=verify)
    else:
        client = api_client

    url = client.wml_credentials['url']

    params = client._params(skip_for_create=True)
    if limit < 1 or limit > 200:
        raise InvalidValue(value_name="limit", reason=f"The given value {limit} is not in the range <1, 200>")
    else:
        params.update({'limit': limit})

    if asynchronous or get_all:
        resource_generator = NextResourceGenerator(client,
                                                   url=url,
                                                   href="ml/v4/custom_foundation_models",
                                                   params=params,
                                                   _all=get_all)

        if asynchronous:
            return resource_generator

        resources = []
        for entry in resource_generator:
            resources.extend(entry['resources'])
        return {
            "resources": resources
        }

    response = requests.get(f"{url}/ml/v4/custom_foundation_models", 
                            params=params,
                            headers=client._get_headers()) 
    if response.status_code == 200:
        if model_id:
            resources = [res for res in response.json()['resources'] if res['model_id'] == model_id]
            return resources[0] if resources else {}
        else:
            return response.json()
    elif response.status_code == 404:
        raise WMLClientError(Messages.get_message(url, message_id="custom_models_no_model_specs"), url)
    else:
        msg = f"Getting failed. Reason: {response.text}"
        raise WMLClientError(msg)


def get_model_lifecycle(url: str, model_id: str) -> Union[list, None]:
    """
    Operation to retrieve the list of model lifecycle data.

    :param url: environment url
    :type url: str

    :param model_id: the type of model to use
    :type model_id: str

    :return: list of deployed foundation model lifecycle data
    :rtype: list

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import get_model_lifecycle

        get_model_lifecycle(
            url="https://us-south.ml.cloud.ibm.com",
            model_id="ibm/granite-13b-instruct-v2"
            )
    """
    model_specs = get_model_specs(url=url)
    model_spec = next((model_metadata for model_metadata in model_specs.get('resources', [])
                       if model_metadata.get('model_id') == model_id), None)
    return model_spec.get('lifecycle') if model_spec is not None else None


def _check_model_state(url, model_id):

    default_warning_template = (
        "Model '{model_id}' is in {state} state from {start_date} until {withdrawn_start_date}. "
        "IDs of alternative models: {alternative_model_ids}. "
        "Further details: https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-lifecycle.html?context=wx&audience=wdp")

    lifecycle = get_model_lifecycle(url, model_id)

    modes_list = [ids.get("id") for ids in (lifecycle or [])]
    deprecated_or_constricted_warning_template_cpd = (
            "Model '{model_id}' is in {state} state. "
            "IDs of alternative models: {alternative_model_ids}. "
            "Further details: https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-lifecycle.html?context=wx&audience=wdp")

    if lifecycle and SpecStates.DEPRECATED.value in modes_list:
        model_lifecycle = next((el for el in lifecycle if el.get('id') == SpecStates.DEPRECATED.value), None)
        if model_lifecycle.get('since_version'):
            warnings.warn(deprecated_or_constricted_warning_template_cpd.format(
                model_id=model_id,
                state=(model_lifecycle.get('label') or SpecStates.DEPRECATED.value),
                alternative_model_ids=', '.join(model_lifecycle.get('alternative_model_ids', ["None"]))
            ), category=LifecycleWarning)
        else:
            warnings.warn(default_warning_template.format(
                model_id=model_id,
                state=(model_lifecycle.get('label') or SpecStates.DEPRECATED.value),
                start_date=model_lifecycle.get('start_date'),
                withdrawn_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.WITHDRAWN.value), None),
                alternative_model_ids=', '.join(model_lifecycle.get('alternative_model_ids', ["None"]))
            ), category=LifecycleWarning)

    elif lifecycle and SpecStates.CONSTRICTED.value in modes_list:
        model_lifecycle = next((el for el in lifecycle if el.get('id') == SpecStates.CONSTRICTED.value), None)
        if model_lifecycle.get('since_version'):
            warnings.warn(deprecated_or_constricted_warning_template_cpd.format(
                model_id=model_id,
                state=(model_lifecycle.get('label') or SpecStates.CONSTRICTED.value),
                alternative_model_ids=', '.join(model_lifecycle.get('alternative_model_ids', ["None"]))
            ), category=LifecycleWarning)
        else:
            warnings.warn(default_warning_template.format(
                model_id=model_id,
                state=(model_lifecycle.get('label') or SpecStates.CONSTRICTED.value),
                start_date=model_lifecycle.get('start_date'),
                withdrawn_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.WITHDRAWN.value), None),
                alternative_model_ids=', '.join(model_lifecycle.get('alternative_model_ids', ["None"]))
            ), category=LifecycleWarning)


def get_model_specs_with_prompt_tuning_support(url: str) -> dict:
    """
    Operations to query the details of the deployed foundation models with prompt tuning support.

    :param url: environment url
    :type url: str

    :return: list of deployed foundation model specs with prompt tuning support
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import get_model_specs_with_prompt_tuning_support

        get_model_specs_with_prompt_tuning_support(
            url="https://us-south.ml.cloud.ibm.com"
            )
    """
    try:
        try:
            return _get_foundation_models_spec(url=f"{url}/ml/v1/foundation_model_specs",
                                               operation_name="Get available foundation models",
                                               additional_params={"filters": "function_prompt_tune_trainable"})
        except WMLClientError:  # Remove on CPD 5.0 release
            return _get_foundation_models_spec(url=f"{url}/ml/v1-beta/foundation_model_specs",
                                               operation_name="Get available foundation models",
                                               additional_params={"filters": "function_prompt_tune_trainable"})
    except WMLClientError as e:
        raise WMLClientError(Messages.get_message(url, message_id="fm_prompt_tuning_no_model_specs"), e)


def get_supported_tasks(url: str) -> dict:
    """
    Operation to retrieve the list of tasks that are supported by the foundation models.

    :param url: environment url
    :type url: str

    :return: list of tasks that are supported by the foundation models
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import get_supported_tasks

        get_supported_tasks(
            url="https://us-south.ml.cloud.ibm.com"
            )
    """
    try:
        try:
            return _get_foundation_models_spec(f"{url}/ml/v1/foundation_model_tasks",
                                               "Get tasks that are supported by the foundation models.")
        except WMLClientError:  # Remove on CPD 5.0 release
            return _get_foundation_models_spec(f"{url}/ml/v1-beta/foundation_model_tasks",
                                               "Get tasks that are supported by the foundation models.")
    except WMLClientError as e:
        raise WMLClientError(Messages.get_message(url, message_id="fm_prompt_tuning_no_supported_tasks"), e)


def get_all_supported_tasks_dict(url="https://us-south.ml.cloud.ibm.com") -> dict:
    tasks_dict = dict()
    for task_spec in get_supported_tasks(url).get('resources', []):
        tasks_dict[task_spec['label'].replace("-", "_").replace(" ", "_").upper()] = task_spec['task_id']
    return tasks_dict


def load_request_json(run_id, wml_client, run_params = None) -> dict:
    if run_params is None:
        run_params = wml_client.training.get_details(run_id)

    model_request_path = run_params['entity'].get('results_reference', {}).get('location', {}).get('model_request_path')

    if model_request_path is None:
        raise WMLClientError("Missing model_request_path in run_params. Verify if the training run has been completed.")

    if wml_client.CLOUD_PLATFORM_SPACES:
        results_reference = DataConnection._from_dict(run_params['entity']['results_reference'])

        if run_params['entity']['results_reference']['type'] == DataConnectionTypes.CA:
            results_reference.location.file_name = model_request_path
        else:
            results_reference.location.path = model_request_path

        results_reference.set_client(wml_client)
        request_json_bytes = results_reference.read(raw=True, binary=True)

        # download from cos

    elif wml_client.CPD_version >= 4.8:
        asset_parts = model_request_path.split('/')
        model_request_asset_url = '/'.join(asset_parts[asset_parts.index('assets') + 1:])
        request_json_bytes = load_file_from_file_system_nonautoai(wml_client=wml_client,
                                                                  file_path=model_request_asset_url).read()
    else:
        raise WMLClientError("Unsupported environment for this action")

    return json_loads(request_json_bytes.decode())


def is_training_prompt_tuning(training_id, wml_client):
    """Returns True if training_id is connected to prompt tuning"""
    if training_id is None:
        return False
    run_params = wml_client.training.get_details(training_uid=training_id)
    return bool(run_params['entity'].get('prompt_tuning'))


class TemplateFormatter(Formatter):
    def check_unused_args(self, 
                          used_args: List[(int | str)], 
                          args: Sequence, 
                          kwargs: Dict[str, Any]) -> None:
        """Check for unused args."""
        extra_args = set(kwargs).difference(used_args)
        if extra_args:
            raise KeyError(extra_args)


class HAPDetectionWarning(UserWarning): ...


class PIIDetectionWarning(UserWarning): ...


class LifecycleWarning(UserWarning): ...


class WatsonxLLMDeprecationWarning(UserWarning): ...


def _raise_watsonxllm_deprecation_warning() -> None:
    warnings.warn(f"ibm_watson_machine_learning.foundation_models.extensions.langchain.WatsonxLLM"
                  f" is deprecated and will not be supported in the future. "
                  f"Please import from langchain-ibm instead.\n"
                  "To install langchain-ibm run `pip install -U langchain-ibm`.",
                  category=WatsonxLLMDeprecationWarning,
                  stacklevel=2
                  )



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import abc
import logging

import numpy

from ibmfl.data.data_util import (
    get_max,
    get_mean,
    get_min,
    get_minmaxscaler,
    get_normalizer,
    get_quantile,
    get_standardscaler,
    get_std,
    get_var,
)
from ibmfl.exceptions import FLException

logger = logging.getLogger(__name__)


class DataHandler(abc.ABC):
    """
    Base class to load and pre-process data.
    """

    def __init__(self, **kwargs):
        self.x_train = None
        self.y_train = None
        self.x_test = None
        self.y_test = None
        self.preprocessor = None
        self.testing_generator = None
        self.training_generator = None

    @abc.abstractmethod
    def get_data(self, **kwargs):
        """
        Access the local dataset and return the training and testing dataset
        as a tuple.

        :param kwargs:
        :return: `tuple`. (training_set, testing_set)
        """
        raise NotImplementedError

    def get_train_counts(self):
        """
        Returns the training sample size.

        :return: The training sample size
        :rtype: `int`
        """
        if self.x_train is not None and isinstance(self.x_train, numpy.ndarray):
            return self.x_train.shape[0]
        else:
            logger.info(
                "The attribute x_train is None or is not of numpy.ndarray format!"
                "Trying to access training counts assuming data is loaded \
                as DataGenerator type."
            )
            try:
                counts = len(self.training_generator.filenames)
                assert isinstance(counts, int)
                return counts
            except:
                raise FLException("Error ocurred during accessing the training sample size.")

    def get_val_data(self, **kwargs):
        """
        Access the local dataset and return the validation dataset
        as a tuple.
        :param kwargs:
        :return: `tuple`. (validation_set)
        """
        pass

    def get_statistics_of_training_data(self, sample_data_schema, lst_stats_name, **kwargs):
        """
        Return the corresponding statistics, which is specified by the
        provided list of statistics names, of the local training dataset.

        :param sample_data_schema: Provided data with only feature values. \
        Assuming the dataset has shape (num_samples, num_features).
        :type sample_data_schema: `np.array`
        :param lst_stats_name: A list of statistics names, \
        all in lowercase form, for example, ['min'], ['mean', 'variance'], etc.
        :type lst_stats_name: `list` of `str`
        :param kwargs: Additional parameters to obtain the statistics.
        :type kwargs: `dict`
        :return: The requested statistics based on the local dataset.
        :rtype: `dict`
        """
        if sample_data_schema is None and self.x_train is None:
            raise FLException("No data is provided!")
        elif sample_data_schema is None and self.x_train is not None:
            sample_data_schema = self.x_train
            logger.warning(
                "No dataset is provided, use the data schema from " "training dataset(x_train) as the default one."
            )

        if type(lst_stats_name) is not list:
            raise FLException(
                "list of requested statistics badly form. "
                "It should of type list, but it is instead "
                "of type " + str(type(lst_stats_name))
            )
        list_stats = {}
        if not isinstance(sample_data_schema, numpy.ndarray):
            raise FLException(
                "Expecting the local dataset to be of type "
                "numpy.ndarray, instead it is of type " + str(type(sample_data_schema))
            )

        while len(lst_stats_name) != 0:
            tmp_stat_name = lst_stats_name.pop()
            if type(tmp_stat_name) is not str:
                logger.warning(
                    "Skipping the current requested statistics "
                    "name. It should be of type string, "
                    "but now it is instead of type" + str(type(tmp_stat_name))
                )
            elif tmp_stat_name == "min" or tmp_stat_name == "minimum":
                list_stats[tmp_stat_name] = get_min(sample_data_schema)
            elif tmp_stat_name == "max" or tmp_stat_name == "maximum":
                list_stats[tmp_stat_name] = get_max(sample_data_schema)
            elif tmp_stat_name == "mean":
                list_stats[tmp_stat_name] = get_mean(sample_data_schema)
            elif tmp_stat_name == "var" or tmp_stat_name == "variance":
                list_stats[tmp_stat_name] = get_var(sample_data_schema)
            elif tmp_stat_name == "std" or tmp_stat_name == "standard deviation":
                list_stats[tmp_stat_name] = get_std(sample_data_schema)
            elif tmp_stat_name == "quantile":
                if kwargs and "q" in kwargs:
                    list_stats[tmp_stat_name] = get_quantile(sample_data_schema, percentage=kwargs["q"])
                else:
                    raise FLException("Cannot compute quantile, " "missing quantile requirement.")
            else:
                logger.warning("Current required statistics " "is not supported. Skipping...")

        return list_stats

    def get_preprocessor(self, sample_data_schema, preprocessor_name, **kwargs):
        """
        Set the data preprocessor of the data handler class as the requested
        type of preprocessor. The supported preprocessors
        include `normalizer`, `standardscaler` and `minmaxscaler`.
        All provided based on `sklearn.preprocessing` module.
        The preprocessor can be applied to perform the
        required preprocessing step for the party's local dataset
        via `transform` method.

        :param sample_data_schema: Provided data with only feature values to \
        initialize the preprocessor. \
        Assuming the dataset has shape (num_samples, num_features).
        :type sample_data_schema: `np.array`
        :param preprocessor_name: The requested preprocessor name in lowercase.
        :type preprocessor_name: `str`
        :param kwargs: Additional parameters to obtain the preprocessor.
        :type kwargs: `dict`
        :return: None
        """
        if sample_data_schema is None and self.x_train is None:
            raise FLException("No data is provided!")
        elif sample_data_schema is None and self.x_train is not None:
            sample_data_schema = self.x_train
            logger.warning(
                "No dataset is provided, use the data schema from " "training dataset(x_train) as the default one."
            )

        if type(preprocessor_name) is not str:
            raise FLException(
                "Expecting the requested preprocessor "
                "to be of type string, instead it is of type" + str(type(preprocessor_name))
            )

        if not isinstance(sample_data_schema, numpy.ndarray):
            raise FLException(
                "Expecting the local dataset to be of type "
                "numpy.ndarray, instead it is of type " + str(type(sample_data_schema))
            )

        if preprocessor_name == "normalizer" or preprocessor_name == "normalization":
            if "norm" in kwargs:
                self.preprocessor = get_normalizer(sample_data_schema, norm=kwargs["norm"])
            else:
                self.preprocessor = get_normalizer(sample_data_schema)
        elif preprocessor_name == "standardscaler" or preprocessor_name == "standardization":
            mean_val = None
            std = None
            if "mean" in kwargs:
                mean_val = kwargs["mean"]
            if "scale" in kwargs:
                std = kwargs["scale"]
            self.preprocessor = get_standardscaler(sample_data_schema, mean_val=mean_val, std=std)
        elif preprocessor_name == "minmaxscaler":
            if "feature_range" in kwargs:
                self.preprocessor = get_minmaxscaler(sample_data_schema, feature_range=kwargs["feature_range"])
            else:
                self.preprocessor = get_minmaxscaler(sample_data_schema)
        else:
            logger.warning("Required preprocessor is not supported. " "Skipping...")



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------


import logging

from ibmfl.util.serialization.serializer import Serializer

import msgpack

from sklearn.ensemble._hist_gradient_boosting.predictor import TreePredictor

from ddsketch.ddsketch import DDSketch
from ddsketch.pb.proto import DDSketchProto
from google.protobuf.json_format import MessageToJson,Parse

from skops.io import dumps, loads

import numpy as np

from enum import Enum

logger = logging.getLogger(__name__)

class Serializers(Enum):
    Sketch = 1
    Predictor = 2
    DType = 3



#
# DDSketch Serializer
#
class SketchSerializer(Serializer):

    @staticmethod
    def serialize(x):
        """
        Serialize 
        """
        proto = DDSketchProto.to_proto(x)
        json = MessageToJson(proto,preserving_proto_field_name=True)
        return msgpack.packb(json)

    @staticmethod
    def deserialize(serialization):
        """
        Deserialize 

        :param serialization: Serialization of the object
        :return: deserialized object
        """
        unpacked = msgpack.unpackb(serialization)
        proto = Parse(unpacked, DDSketchProto.to_proto(DDSketch()))
        sketch = DDSketchProto.from_proto(proto)
        return sketch
        
#
# TreePredictor Serializer
#
class PredictorSerializer(Serializer):

    @staticmethod
    def serialize(x):
        """
        Serialize 
        """
        
        return dumps(x)

    @staticmethod
    def deserialize(serialization):
        """
        Deserialize 

        :param serialization: Serialization of the object
        :return: deserialized object
        """
        t = loads(serialization, trusted = ['sklearn.ensemble._hist_gradient_boosting.predictor.TreePredictor'])

        return t
        

#
# Numpy DType Serializer
#
class DTypeSerializer(Serializer):

    @staticmethod
    def serialize(x):
        """
        Serialize 
        """
        
        # Take typestr
        return bytes(x.descr[0][1],'utf-8')

    @staticmethod
    def deserialize(serialization):
        """
        Deserialize 

        :param serialization: Serialization of the object
        :return: deserialized object
        """

        # typestr to dtype
        t = np.dtype(str(serialization,'utf-8'))

        return t
        



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------


from ibmfl.exceptions import SerializationException
import logging

import msgpack
import msgpack_numpy as m
import numpy as np


from ibmfl.util.serialization.serializers import SketchSerializer, PredictorSerializer, DTypeSerializer
from ibmfl.util.serialization.serializers import Serializers

from sklearn.ensemble._hist_gradient_boosting.predictor import TreePredictor

from ddsketch.ddsketch import DDSketch

logger = logging.getLogger(__name__)


def encode(obj):
    if isinstance(obj, DDSketch):
        return msgpack.ExtType(Serializers.Sketch.value,SketchSerializer.serialize(obj))
    elif isinstance(obj, TreePredictor):
        return msgpack.ExtType(Serializers.Predictor.value,PredictorSerializer.serialize(obj))
    elif isinstance(obj, np.dtype):
        return msgpack.ExtType(Serializers.DType.value, DTypeSerializer.serialize(obj))
    else :
        return m.encode(obj)
    

def decode(obj, chain=None):
    return m.decode(obj)

def ext_hook(code, data):
    if code == Serializers.Sketch.value:
        return SketchSerializer.deserialize(data)
    elif code == Serializers.Predictor.value:
        return PredictorSerializer.deserialize(data)
    elif code == Serializers.DType.value:
        return DTypeSerializer.deserialize(data)
    return msgpack.ExtType(code.data)

def pack(x):
    try:
        return msgpack.packb(x, default=encode)
    except TypeError :
        logger.error("Failed to serialize object",exc_info=True)
        raise SerializationException("Unable to serialize object")

def unpack(x_enc):
    try:
        return msgpack.unpackb(x_enc, object_hook=decode, ext_hook=ext_hook, strict_map_key=False)
    except TypeError :
        logger.error("Failed to serialize {}".format(x_enc))
        raise SerializationException("Unable to deserialize object")
        





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import copy
from typing import List, Union
from warnings import warn

from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.utils.autoai.enums import (
    TShirtSize, ClassificationAlgorithms, RegressionAlgorithms, ForecastingAlgorithms, PredictionType, Metrics, \
    Transformers, DataConnectionTypes, PipelineTypes, PositiveLabelClass, ClassificationAlgorithmsCP4D,
    RegressionAlgorithmsCP4D, ForecastingAlgorithmsCP4D, SamplingTypes, ImputationStrategy, ForecastingPipelineTypes,
    TimeseriesAnomalyPredictionPipelineTypes, TimeseriesAnomalyPredictionAlgorithms)
from ibm_watson_machine_learning.utils.autoai.errors import LocalInstanceButRemoteParameter, MissingPositiveLabel, \
    NonForecastPredictionColumnMissing, ForecastPredictionColumnsMissing, ForecastingCannotBeRunAsLocalScenario, \
    TSNotSupported, TSADNotSupported, ImputationListNotSupported, \
    MissingEstimatorForExistingBatchedEstimator, TimeseriesAnomalyPredictionFeatureColumnsMissing, \
    TimeseriesAnomalyPredictionCannotBeRunAsLocalScenario, TimeseriesAnomalyPredictionUnsupportedMetric
from ibm_watson_machine_learning.utils.autoai.utils import check_dependencies_versions, \
    validate_additional_params_for_optimizer, validate_optimizer_enum_values, \
    translate_imputation_string_strategy_to_enum, translate_estimator_string_to_enum, \
    translate_batched_estimator_string_to_enum
from ibm_watson_machine_learning.workspace import WorkSpace
from ibm_watson_machine_learning.wml_client_error import ForbiddenActionForGitBasedProject, WMLClientError, ParamOutOfRange
from ibm_watson_machine_learning.messages.messages import Messages
from .engines import WMLEngine
from .optimizers import LocalAutoPipelines, RemoteAutoPipelines
from .runs import AutoPipelinesRuns, LocalAutoPipelinesRuns
from ..base_experiment.base_experiment import BaseExperiment

__all__ = [
    "AutoAI"
]


class AutoAI(BaseExperiment):
    """AutoAI class for pipeline models optimization automation.

    :param wml_credentials: credentials to Watson Machine Learning instance
    :type wml_credentials: dict

    :param project_id: ID of the Watson Studio project
    :type project_id: str, optional

    :param space_id: ID of the Watson Studio Space
    :type space_id: str, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.experiment import AutoAI

        experiment = AutoAI(
            wml_credentials={
                "apikey": "...",
                "iam_apikey_description": "...",
                "iam_apikey_name": "...",
                "iam_role_crn": "...",
                "iam_serviceid_crn": "...",
                "instance_id": "...",
                "url": "https://us-south.ml.cloud.ibm.com"
            },
            project_id="...",
            space_id="...")
    """
    # note: initialization of AutoAI enums as class properties

    # note: Enums with estimators can be overwritten  in _init based on environment type (CPD or Cloud)
    ClassificationAlgorithms = ClassificationAlgorithms
    RegressionAlgorithms = RegressionAlgorithms
    ForecastingAlgorithms = ForecastingAlgorithms
    # end note
    TShirtSize = TShirtSize
    PredictionType = PredictionType
    Metrics = Metrics
    Transformers = Transformers
    DataConnectionTypes = DataConnectionTypes
    PipelineTypes = PipelineTypes
    SamplingTypes = SamplingTypes

    def __init__(self,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None,
                 verify=None) -> None:
        # note: as workspace is not clear enough to understand, there is a possibility to use pure
        # wml credentials with project and space IDs, but in addition we
        # leave a possibility to use a previous WorkSpace implementation, it could be passed as a first argument
        if wml_credentials is None:
            self._workspace = None
            self.runs = LocalAutoPipelinesRuns()

        else:
            if isinstance(wml_credentials, WorkSpace):
                self._workspace = wml_credentials
            else:
                self._workspace = WorkSpace(wml_credentials=wml_credentials.copy(),
                                            project_id=project_id,
                                            space_id=space_id,
                                            verify=verify)

            self.project_id = self._workspace.project_id
            self.space_id = self._workspace.space_id
            self.runs = AutoPipelinesRuns(engine=WMLEngine(self._workspace))
            self.runs._workspace = self._workspace

        #self._block_autoai_on_git_based_project()

        self._init_estimator_enums()

        self._20_class_limit_removal_test = False
        # --- end note

    def runs(self, *, filter: str) -> Union['AutoPipelinesRuns', 'LocalAutoPipelinesRuns']:
        """Get the historical runs but with WML Pipeline name filter (for remote scenario).
        Get the historical runs but with experiment name filter (for local scenario).

        :param filter: WML Pipeline name to filter the historical runs or experiment name to filter
            the local historical runs
        :type filter: str

        :return: object managing the list of runs
        :rtype: AutoPipelinesRuns or LocalAutoPipelinesRuns

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI

            experiment = AutoAI(...)
            experiment.runs(filter='Test').list()
        """

        if self._workspace is None:
            return LocalAutoPipelinesRuns(filter=filter)

        else:
            return AutoPipelinesRuns(engine=WMLEngine(self._workspace.wml_client), filter=filter)

    def optimizer(self,
                  name: str,
                  *,
                  prediction_type: 'PredictionType',
                  prediction_column: str = None,
                  prediction_columns: List[str] = None,
                  timestamp_column_name: str = None,
                  scoring: 'Metrics' = None,
                  desc: str = None,
                  test_size: float = None,  # deprecated
                  holdout_size: float = None,
                  max_number_of_estimators: int = None,
                  train_sample_rows_test_size: float = None,
                  include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms',
                                                      'ForecastingAlgorithms', 'TimeseriesAnomalyPredictionAlgorithms']] = None,
                  daub_include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms']] = None,  # deprecated
                  include_batched_ensemble_estimators: List[Union['BatchedClassificationAlgorithms',
                                                                  'BatchedRegressionAlgorithms']] = None,
                  backtest_num: int = None,
                  lookback_window: int = None,
                  forecast_window: int = None,
                  backtest_gap_length: int = None,
                  feature_columns: List[str] = None,
                  pipeline_types: List[Union['ForecastingPipelineTypes', 'TimeseriesAnomalyPredictionPipelineTypes']] = None,
                  supporting_features_at_forecast: bool = None,
                  cognito_transform_names: List['Transformers'] = None,
                  data_join_graph: 'DataJoinGraph' = None,
                  csv_separator: Union[List[str], str] = ',',
                  excel_sheet: Union[str, int] = None,
                  encoding: str = 'utf-8',
                  positive_label: str = None,
                  data_join_only: bool = False,
                  drop_duplicates: bool = True,
                  outliers_columns: list = None,
                  text_processing: bool = None,
                  word2vec_feature_number: int = None,
                  daub_give_priority_to_runtime: float = None,
                  fairness_info: dict = None,
                  sampling_type: 'SamplingTypes' = None,
                  sample_size_limit: int = None,
                  sample_rows_limit: int = None,
                  sample_percentage_limit: float = None,
                  n_parallel_data_connections: int = None,
                  number_of_batch_rows: int = None,
                  categorical_imputation_strategy: ImputationStrategy = None,
                  numerical_imputation_strategy: ImputationStrategy = None,
                  numerical_imputation_value: float = None,
                  imputation_threshold: float = None,
                  retrain_on_holdout: bool = None,
                  categorical_columns: list = None,
                  numerical_columns: list = None,
                  test_data_csv_separator: Union[List[str], str] = ',',
                  test_data_excel_sheet: str = None,
                  test_data_encoding: str = 'utf-8',
                  confidence_level: float = None,
                  incremental_learning: bool = None,
                  early_stop_enabled: bool = None,
                  early_stop_window_size: int = None,
                  time_ordered_data: bool = None,
                  feature_selector_mode: str = None,
                  **kwargs) -> Union['RemoteAutoPipelines', 'LocalAutoPipelines']:
        """
        Initialize an AutoAi optimizer.

        :param name: name for the AutoPipelines
        :type name: str

        :param prediction_type: type of the prediction
        :type prediction_type: PredictionType

        :param prediction_column: name of the target/label column, required for `multiclass`, `binary` and `regression`
            prediction types
        :type prediction_column: str, optional

        :param prediction_columns: names of the target/label columns, required for `forecasting` prediction type
        :type prediction_columns: list[str], optional

        :param timestamp_column_name: name of timestamp column for time series forecasting
        :type timestamp_column_name: str, optional

        :param scoring: type of the metric to optimize with, not used for forecasting
        :type scoring: Metrics, optional

        :param desc: description
        :type desc: str, optional

        :param test_size: deprecated, use `holdout_size` instead

        :param holdout_size: percentage of the entire dataset to leave as a holdout
        :type holdout_size: float, optional

        :param max_number_of_estimators: maximum number (top-K ranked by DAUB model selection)
            of the selected algorithm, or estimator types, for example `LGBMClassifierEstimator`,
            `XGBoostClassifierEstimator`, or `LogisticRegressionEstimator` to use in pipeline composition,
            the default is `None` that means the true default value will be determined by
            the internal different algorithms, where only the highest ranked by model selection algorithm type is used
        :type max_number_of_estimators: int, optional

        :param train_sample_rows_test_size: training data sampling percentage
        :type train_sample_rows_test_size: float, optional

        :param daub_include_only_estimators: deprecated, use `include_only_estimators` instead
        
        :param include_batched_ensemble_estimators: list of batched ensemble estimators to include 
            in computation process, see: AutoAI.BatchedClassificationAlgorithms, AutoAI.BatchedRegressionAlgorithms
        :type include_batched_ensemble_estimators: 
            list[BatchedClassificationAlgorithms or BatchedRegressionAlgorithms], optional

        :param include_only_estimators: list of estimators to include in computation process, see:
            AutoAI.ClassificationAlgorithms, AutoAI.RegressionAlgorithms or AutoAI.ForecastingAlgorithms
        :type include_only_estimators: List[ClassificationAlgorithms or RegressionAlgorithms or ForecastingAlgorithms]], optional

        :param backtest_num: number of backtests used for forecasting prediction type, default value: 4,
            value from range [0, 20]
        :type backtest_num: int, optional

        :param lookback_window: length of lookback window used for forecasting prediction type,
            default value: 10, if set to -1 lookback window will be auto-detected
        :type lookback_window: int, optional

        :param forecast_window: length of forecast window used for forecasting prediction type, default value: 1,
            value from range [1, 60]
        :type forecast_window: int, optional

        :param backtest_gap_length: gap between backtests used for forecasting prediction type,
            default value: 0, value from range [0, data length / 4]
        :type backtest_gap_length: int, optional

        :param feature_columns: list of feature columns used for forecasting prediction type,
            may contain target column and/or supporting feature columns, list of columns to be detected whether there are anomalies for timeseries anomaly prediction type
        :type feature_columns: list[str], optional

        :param pipeline_types: list of pipeline types to be used for forecasting or timeseries anomaly prediction type
        :type pipeline_types: list[ForecastingPipelineTypes or TimeseriesAnomalyPredictionPipelineTypes], optional

        :param supporting_features_at_forecast: enables usage of future supporting feature values during forecast
        :type supporting_features_at_forecast: bool, optional

        :param cognito_transform_names: list of transformers to include in the feature enginnering computation process,
            see: AutoAI.Transformers
        :type cognito_transform_names: list[Transformers], optional

        :param csv_separator: the separator, or list of separators to try for separating columns in a CSV file,
            not used if the file_name is not a CSV file, default is ','
        :type csv_separator: list[str] or str, optional

        :param excel_sheet: name of the excel sheet to use, only applicable when xlsx file is an input,
            support for number of the sheet is deprecated, by default first sheet is used
        :type excel_sheet: str, optional

        :param encoding: encoding type for CSV training file
        :type encoding: str, optional

        :param positive_label: the positive class to report when binary classification, when multiclass or regression,
            this will be ignored
        :type positive_label: str, optional

        :param t_shirt_size: the size of the remote AutoAI POD instance (computing resources),
            only applicable to a remote scenario, see: AutoAI.TShirtSize
        :type t_shirt_size: TShirtSize, optional

        :param data_join_graph: a graph object with definition of join structure for multiple input data sources,
            data preprocess step for multiple files
        :type data_join_graph: DataJoinGraph, optional

        :param data_join_only: if `True` only preprocessing will be executed
        :type data_join_only: bool, optional

        :param drop_duplicates: if `True` duplicated rows in data will be removed before further processing
        :type drop_duplicates: bool, optional

        :param outliers_columns: replace outliers with NaN using IQR method for specified columns. By default,
            turned ON for regression learning_type and target column. To turn OFF empty list of columns must be passed
        :type outliers_columns: list, optional

        :param text_processing: if `True` text processing will be enabled, applicable only on Cloud
        :type text_processing: bool, optional

        :param word2vec_feature_number: number of features which will be generated from text column,
            will be applied only if `text_processing` is `True`, if `None` the default value will be taken
        :type word2vec_feature_number: int, optional

        :param daub_give_priority_to_runtime: the importance of run time over score for pipelines ranking,
            can take values between 0 and 5, if set to 0.0 only score is used,
            if set to 1 equally score and runtime are used, if set to value higher than 1
            the runtime gets higher importance over score
        :type daub_give_priority_to_runtime: float, optional

        :param fairness_info: dictionary that specifies metadata needed for measuring fairness,
            it contains three key values: `favorable_labels`, `unfavorable_labels` and `protected_attributes`,
            the `favorable_labels` attribute indicates that when the class column contains one of the value from list,
            that is considered a positive outcome, the `unfavorable_labels` is oposite to the `favorable_labels`
            and is obligatory for regression learning type, a protected attribute is a list of features that partition
            the population into groups whose outcome should have parity, if protected attribute is empty list
            then automatic detection of protected attributes will be run,
            if `fairness_info` is passed then fairness metric will be calculated
        :type fairness_info: fairness_info

        :param n_parallel_data_connections: number of maximum parallel connection to data source,
            supported only for IBM Cloud Pak® for Data 4.0.1 and above
        :type n_parallel_data_connections: int, optional

        :param categorical_imputation_strategy: missing values imputation strategy for categorical columns

            Possible values (only non-forecasting scenario):

            - ImputationStrategy.MEAN
            - ImputationStrategy.MEDIAN
            - ImputationStrategy.MOST_FREQUENT (default)

        :type categorical_imputation_strategy: ImputationStrategy, optional


        :param numerical_imputation_strategy: missing values imputation strategy for numerical columns

            Possible values (non-forecasting scenario):

            - ImputationStrategy.MEAN
            - ImputationStrategy.MEDIAN (default)
            - ImputationStrategy.MOST_FREQUENT

            Possible values (forecasting scenario):

            - ImputationStrategy.MEAN
            - ImputationStrategy.MEDIAN
            - ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS (default)
            - ImputationStrategy.VALUE
            - ImputationStrategy.FLATTEN_ITERATIVE
            - ImputationStrategy.LINEAR
            - ImputationStrategy.CUBIC
            - ImputationStrategy.PREVIOUS
            - ImputationStrategy.NEXT
            - ImputationStrategy.NO_IMPUTATION

        :param numerical_imputation_value: value for filling missing values if numerical_imputation_strategy
            is set to ImputationStrategy.VALUE, for forecasting only
        :type numerical_imputation_value: float, optional

        :param imputation_threshold: maximum threshold of missing values imputation, for forecasting only
        :type imputation_threshold: float, optional

        :param retrain_on_holdout: if True final pipelines will be train also on holdout data
        :type retrain_on_holdout: bool, optional

        :param categorical_columns: list of columns names that must be treated as categorical
        :type categorical_columns: list, optional

        :param numerical_columns: list of columns names that must be treated as numerical
        :type numerical_columns: list, optional

        :param sampling_type: type of sampling data for training, one of SamplingTypes enum values,
            default is SamplingTypes.FIRST_N_RECORDS, supported only for IBM Cloud Pak® for Data 4.0.1 and above
        :type sampling_type: str, optional

        :param sample_size_limit: the size of sample upper bound (in bytes). The default value is 1GB,
            supported only for IBM Cloud Pak® for Data 4.5 and above
        :type sample_size_limit: int, optional

        :param sample_rows_limit: the size of sample upper bound (in rows),
            supported only for IBM Cloud Pak® for Data 4.6 and above
        :type sample_rows_limit: int, optional

        :param sample_percentage_limit: the size of sample upper bound (as fraction of dataset size),
            supported only for IBM Cloud Pak® for Data 4.6 and above
        :type sample_percentage_limit: float, optional

        :param number_of_batch_rows: number of rows to read in each batch when reading from flight connection
        :type number_of_batch_rows: int, optional

        :param test_data_csv_separator: the separator, or list of separators to try for separating
            columns in a CSV user-defined holdout/test file, not used if the file_name is not a CSV file,
            default is ','
        :type test_data_csv_separator: list[str] or str, optional

        :param test_data_excel_sheet: name of the excel sheet to use for user-defined holdout/test data, 
            only use when xlsx file is an test, dataset file, by default first sheet is used
        :type test_data_excel_sheet: str or int, optional

        :param test_data_encoding: encoding type for CSV user-defined holdout/test file
        :type test_data_encoding: str, optional

        :param confidence_level: when the pipeline "PointwiseBoundedHoltWinters" or "PointwiseBoundedBATS" is used, 
            the prediction interval is calculated at a given confidence_level to decide if a data record 
            is an anomaly or not, optional for timeseries anomaly prediction
        :type confidence_level: float, optional

        :param incremental_learning: triggers incremental learning process for supported pipelines
        :type incremental_learning: bool, optional

        :param early_stop_enabled: enables early stop for incremental learning process
        :type early_stop_enabled: bool, optional

        :param early_stop_window_size: the number of iterations without score improvements before training stop
        :type early_stop_window_size: int, optional

        :param time_ordered_data: defines user preference about time-based analise. If True, the analysis will
            consider the data as time-ordered and time-based. Supported only for regression.
        :type time_ordered_data: bool, optional

        :param feature_selector_mode: defines if feature selector should be triggered ["on", "off", "auto"].
                The "auto" mode analyses the impact of removal of insignificant features. If there is drop in accuracy,
                the PCA is applied to insignificant features. Principal components describing variance in 30% or higher
                are selected in place of insignificant features. The model is evaluated again. If there is still drop
                in accuracy all features are used.
                The "on" mode removes all insignificant features (0.0. importance). Feature selector is applied during
                cognito phase (applicable to pipelines with feature engineering stage).
        :type feature_selector_mode: str, optional

        :return: RemoteAutoPipelines or LocalAutoPipelines, depends on how you initialize the AutoAI object
        :rtype: RemoteAutoPipelines or LocalAutoPipelines

        **Examples**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(...)

            fairness_info = {
                       "protected_attributes": [
                           {"feature": "Sex", "reference_group": ['male'], "monitored_group": ['female']},
                           {"feature": "Age", "reference_group": [[50,60]], "monitored_group": [[18, 49]]}
                       ],
                       "favorable_labels": ["No Risk"],
                       "unfavorable_labels": ["Risk"],
                       }

            optimizer = experiment.optimizer(
                   name="name of the optimizer.",
                   prediction_type=AutoAI.PredictionType.BINARY,
                   prediction_column="y",
                   scoring=AutoAI.Metrics.ROC_AUC_SCORE,
                   desc="Some description.",
                   holdout_size=0.1,
                   max_number_of_estimators=1,
                   fairness_info= fairness_info,
                   cognito_transform_names=[AutoAI.Transformers.SUM,AutoAI.Transformers.MAX],
                   train_sample_rows_test_size=1,
                   include_only_estimators=[AutoAI.ClassificationAlgorithms.LGBM, AutoAI.ClassificationAlgorithms.XGB],
                   t_shirt_size=AutoAI.TShirtSize.L
               )

            optimizer = experiment.optimizer(
                   name="name of the optimizer.",
                   prediction_type=AutoAI.PredictionType.MULTICLASS,
                   prediction_column="y",
                   scoring=AutoAI.Metrics.ROC_AUC_SCORE,
                   desc="Some description.",
               )
        """
        # note: convert `timeseries` type to PredictionType.FORECASTING:
        if prediction_type == 'timeseries':
            prediction_type = PredictionType.FORECASTING


        if prediction_type != PredictionType.FORECASTING and retrain_on_holdout is None:
            retrain_on_holdout = True

        # Deprecation of excel_sheet as number:
        if isinstance(excel_sheet, int):
            warn(
                message="Support for excel sheet as number of the sheet (int) is deprecated! Please set excel sheet with name of the sheet.")

        if data_join_graph is not None:
            if self._workspace.wml_client.CPD_version >= 4.6:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cpd"))
            elif self._workspace.wml_client.ICP:
                print(Messages.get_message(message_id="obm_deprecation_message_cpd"))
            else:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cloud"))

        if prediction_type == PredictionType.FORECASTING and self._workspace.wml_client.ICP and \
                (self._workspace.wml_client.wml_credentials['version'].startswith('2.5') or \
                        self._workspace.wml_client.wml_credentials['version'].startswith('3.0') or \
                        self._workspace.wml_client.wml_credentials['version'].startswith('3.5')):
            raise TSNotSupported()

        if prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION and self._workspace.wml_client.ICP and \
                self._workspace.wml_client.wml_credentials['version'].startswith(('2.5', '3.0', '3.5', '4.0', '4.5', '4.6')):
            raise TSADNotSupported()

        if prediction_type in (PredictionType.FORECASTING, 'timeseries'):
            if not numerical_imputation_strategy and type(numerical_imputation_strategy) is not list:
                numerical_imputation_strategy = ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS
            elif not numerical_imputation_strategy and type(numerical_imputation_strategy) is list:
                numerical_imputation_strategy = ImputationStrategy.NO_IMPUTATION

            if prediction_column is not None or prediction_columns is None:
                raise ForecastPredictionColumnsMissing()
        elif prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION:
            if feature_columns is None or prediction_column is not None or prediction_columns is not None:
                raise TimeseriesAnomalyPredictionFeatureColumnsMissing()
            if scoring is not None and scoring not in (
                    Metrics.F1_SCORE, Metrics.ROC_AUC_SCORE, Metrics.AVERAGE_PRECISION_SCORE, Metrics.PRECISION_SCORE,
                    Metrics.RECALL_SCORE):
                raise TimeseriesAnomalyPredictionUnsupportedMetric(scoring)
        else:
            if prediction_column is None or prediction_columns is not None:
                raise NonForecastPredictionColumnMissing(prediction_type)

        if test_size:
            print('Note: Using `test_size` is deprecated. Use `holdout_size` instead.')
            if not holdout_size:
                holdout_size = test_size
            test_size = None

        if daub_include_only_estimators:
            print('Note: Using `daub_include_only_estimators` is deprecated. Use `include_only_estimators` instead.')
            if not include_only_estimators:
                include_only_estimators = daub_include_only_estimators
            daub_include_only_estimators = None

        if train_sample_rows_test_size and  self._workspace.wml_client.CPD_version >= 4.6:
            print('Note: Using `train_sample_rows_test_size` is deprecated.'
                  'Use either `sample_rows_limit` or `sample_percentage_limit` instead.')
            if not sample_rows_limit and not sample_percentage_limit:
                if type(train_sample_rows_test_size) is float and train_sample_rows_test_size <= 1:
                    print('Value of `train_sample_rows_test_size` parameter'
                          'will be passed as `sample_percentage_limit`.')
                    sample_percentage_limit = train_sample_rows_test_size
                elif int(train_sample_rows_test_size) == train_sample_rows_test_size and train_sample_rows_test_size > 1:
                    print('Value of `train_sample_rows_test_size` parameter'
                          'will be passed as `sample_rows_limit`.')
                    sample_rows_limit = int(train_sample_rows_test_size)
                train_sample_rows_test_size = None
            elif sample_rows_limit or sample_percentage_limit:
                print('Parameter `train_sample_rows_test_size` will be ignored.')
                train_sample_rows_test_size = None

        def translate_str_imputation_param(x):
            if type(x) is list and prediction_type != PredictionType.FORECASTING:
                raise ImputationListNotSupported()

            if type(x) == str or (type(x) == list and type(x[0]) == str):
                return translate_imputation_string_strategy_to_enum(x, prediction_type)
            else:
                return x

        def translate_str_include_only_estimators_param(x):
            return [translate_estimator_string_to_enum(estimator) for estimator in x]

        def translate_str_include_batched_ensemble_estimators_param(x):
            return [translate_batched_estimator_string_to_enum(estimator) for estimator in x]

        def translate_str_pipeline_types_param(x):
            if prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION:
                return [TimeseriesAnomalyPredictionPipelineTypes(pipeline_type) for pipeline_type in x]
            else:
                return [ForecastingPipelineTypes(pipeline_type) for pipeline_type in x]

        categorical_imputation_strategy = translate_str_imputation_param(categorical_imputation_strategy)
        numerical_imputation_strategy = translate_str_imputation_param(numerical_imputation_strategy)
        include_only_estimators = translate_str_include_only_estimators_param(include_only_estimators) if include_only_estimators else None
        include_batched_ensemble_estimators = translate_str_include_batched_ensemble_estimators_param(include_batched_ensemble_estimators) if include_batched_ensemble_estimators else None
        pipeline_types = translate_str_pipeline_types_param(pipeline_types) if pipeline_types != None else None

        if include_batched_ensemble_estimators:
            for batched_estimator in include_batched_ensemble_estimators:
                basic_estimator_str = batched_estimator.value.split("(")[1][:-1]
                basic_estimator = translate_estimator_string_to_enum(basic_estimator_str)
                if include_only_estimators is not None and basic_estimator not in include_only_estimators:
                    raise MissingEstimatorForExistingBatchedEstimator(batched_estimator, basic_estimator)

        validate_optimizer_enum_values(
            prediction_type=prediction_type,
            daub_include_only_estimators=daub_include_only_estimators,
            include_only_estimators=include_only_estimators,
            include_batched_ensemble_estimators=include_batched_ensemble_estimators,
            cognito_transform_names=cognito_transform_names,
            imputation_strategies=[x for y in list(filter(None, [categorical_imputation_strategy, numerical_imputation_strategy])) for x in (y if type(y) is list else [y])],
            scoring=scoring,
            t_shirt_size=kwargs.get("t_shirt_size", TShirtSize.M),
            is_cpd=self._workspace.wml_client.ICP
        )

        if daub_give_priority_to_runtime is not None:
            if daub_give_priority_to_runtime < 0.0 or daub_give_priority_to_runtime > 5.0:
                raise ParamOutOfRange('daub_give_priority_to_runtime', daub_give_priority_to_runtime, 0.0, 5.0)

        if data_join_graph:
            data_join_graph.problem_type = prediction_type
            data_join_graph.target_column = prediction_column

        if (prediction_type == PredictionType.BINARY and scoring in vars(PositiveLabelClass).values()
                and positive_label is None):
            raise MissingPositiveLabel(scoring, reason=f"\"{scoring}\" needs a \"positive_label\" "
                                                       f"parameter to be defined when used with binary classification.")

        if self._workspace is None and kwargs.get('t_shirt_size'):
            raise LocalInstanceButRemoteParameter(
                "t_shirt_size",
                reason="During LocalOptimizer initialization, \"t_shirt_size\" parameter was provided. "
                       "\"t_shirt_size\" parameter is only applicable to the RemoteOptimizer instance."
            )
        elif self._workspace is None:
            if prediction_type == PredictionType.FORECASTING:
                raise ForecastingCannotBeRunAsLocalScenario()
            if prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION:
                raise TimeseriesAnomalyPredictionCannotBeRunAsLocalScenario()

            reduced_kwargs = copy.copy(kwargs)

            for n in ['_force_local_scenario']:
                if n in reduced_kwargs:
                    del reduced_kwargs[n]

            validate_additional_params_for_optimizer(reduced_kwargs)

            return LocalAutoPipelines(
                name=name,
                prediction_type='classification' if prediction_type in ['binary', 'multiclass'] else prediction_type,
                prediction_column=prediction_column,
                scoring=scoring,
                desc=desc,
                holdout_size=holdout_size,
                max_num_daub_ensembles=max_number_of_estimators,
                train_sample_rows_test_size=train_sample_rows_test_size,
                include_only_estimators=include_only_estimators,
                include_batched_ensemble_estimators=include_batched_ensemble_estimators,
                cognito_transform_names=cognito_transform_names,
                positive_label=positive_label,
                _force_local_scenario=kwargs.get('_force_local_scenario', False),
                **reduced_kwargs
            )

        else:
            reduced_kwargs = copy.copy(kwargs)

            if max_number_of_estimators is None:
                max_number_of_estimators = reduced_kwargs.get('max_num_daub_ensembles', None)

            for n in ['t_shirt_size', 'notebooks', 'autoai_pod_version', 'obm_pod_version', 'max_num_daub_ensembles']:
                if n in reduced_kwargs:
                    del reduced_kwargs[n]

            validate_additional_params_for_optimizer(reduced_kwargs)

            engine = WMLEngine(self._workspace)

            if self._20_class_limit_removal_test:
                engine._20_class_limit_removal_test = True

            optimizer = RemoteAutoPipelines(
                name=name,
                prediction_type=prediction_type,
                prediction_column=prediction_column,
                prediction_columns=prediction_columns,
                timestamp_column_name=timestamp_column_name,
                scoring=scoring,
                desc=desc,
                holdout_size=holdout_size,
                max_num_daub_ensembles=max_number_of_estimators,
                t_shirt_size=self._workspace.restrict_pod_size(t_shirt_size=kwargs.get(
                    't_shirt_size', TShirtSize.M if self._workspace.wml_client.ICP else TShirtSize.L)
                ),
                train_sample_rows_test_size=train_sample_rows_test_size,
                include_only_estimators=include_only_estimators,
                include_batched_ensemble_estimators=include_batched_ensemble_estimators,
                backtest_num=backtest_num,
                lookback_window=lookback_window,
                forecast_window=forecast_window,
                backtest_gap_length=backtest_gap_length,
                cognito_transform_names=cognito_transform_names,
                data_join_graph=data_join_graph,
                drop_duplicates=drop_duplicates,
                outliers_columns=outliers_columns,
                text_processing=text_processing,
                word2vec_feature_number=word2vec_feature_number,
                csv_separator=csv_separator,
                excel_sheet=excel_sheet,
                encoding=encoding,
                positive_label=positive_label,
                data_join_only=data_join_only,
                engine=engine,
                daub_give_priority_to_runtime=daub_give_priority_to_runtime,
                notebooks=kwargs.get('notebooks', True),
                autoai_pod_version=kwargs.get('autoai_pod_version', None),
                obm_pod_version=kwargs.get('obm_pod_version', None),
                fairness_info=fairness_info,
                sampling_type=sampling_type,
                sample_size_limit=sample_size_limit,
                sample_rows_limit=sample_rows_limit,
                sample_percentage_limit=sample_percentage_limit,
                number_of_batch_rows=number_of_batch_rows,
                n_parallel_data_connections=n_parallel_data_connections,
                categorical_imputation_strategy=categorical_imputation_strategy,
                numerical_imputation_strategy=numerical_imputation_strategy,
                numerical_imputation_value=numerical_imputation_value,
                imputation_threshold=imputation_threshold,
                retrain_on_holdout=retrain_on_holdout,
                numerical_columns=numerical_columns,
                categorical_columns=categorical_columns,
                feature_columns=feature_columns,
                pipeline_types=pipeline_types,
                supporting_features_at_forecast=supporting_features_at_forecast,
                test_data_csv_separator=test_data_csv_separator,
                test_data_excel_sheet=test_data_excel_sheet,
                test_data_encoding=test_data_encoding,
                confidence_level=confidence_level,
                incremental_learning=incremental_learning,
                early_stop_enabled=early_stop_enabled,
                early_stop_window_size=early_stop_window_size,
                time_ordered_data=time_ordered_data,
                feature_selector_mode=feature_selector_mode,
                **reduced_kwargs
            )
            optimizer._workspace = self._workspace
            return optimizer

    def _init_estimator_enums(self):
        if self._workspace and self._workspace.wml_client.ICP:
            self.ClassificationAlgorithms = ClassificationAlgorithmsCP4D
            self.RegressionAlgorithms = RegressionAlgorithmsCP4D
            self.ForecastingAlgorithms = ForecastingAlgorithmsCP4D
        else:
            self.ClassificationAlgorithms = ClassificationAlgorithms
            self.RegressionAlgorithms = RegressionAlgorithms
            self.ForecastingAlgorithms = ForecastingAlgorithms

    def _block_autoai_on_git_based_project(self):
        """Raises ForbiddenActionForGitBasedProject error for AutoAI experiments on git based project.
        It can be disabled by setting environment variable ENABLE_AUTOAI to 'true'
        """
        from os import environ

        if self._workspace:
            if getattr(self._workspace.wml_client, 'project_type', None) == 'local_git_storage' \
                    and environ.get('ENABLE_AUTOAI', 'false').lower() == 'false':
                raise ForbiddenActionForGitBasedProject(reason="Creating AutoAI experiment is not supported for git based project.")





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
from enum import Enum


class SpecStates(Enum):
    AVAILABLE = "available"
    SUPPORTED = 'supported'
    DEPRECATED = 'deprecated'
    CONSTRICTED = 'constricted'
    RETIRED = 'retired'
    WITHDRAWN = "withdrawn"

    UNSUPPORTED = 'unsupported'
    CREATE_UNSUPPORTED = 'create-unsupported'



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging

import numpy as np
from diffprivlib.models import GaussianNB

from ibmfl.exceptions import ModelException, ModelInitializationException
from ibmfl.model.model_update import ModelUpdate
from ibmfl.model.sklearn_fl_model import SklearnFLModel
from ibmfl.util.persistence.sklearn import SKLearnPersistence

logger = logging.getLogger(__name__)


class NaiveBayesFLModel(SklearnFLModel):
    """
    Wrapper class for diffprivlib.models.GaussianNB classifier, which itself
    is an implementation of sklearn.naive_bayes.GaussianNB with differential
    privacy.
    """

    def __init__(self, model_name, model_spec, model=None, **kwargs):
        """
        Create a `NaiveBayesFLModel` instance from a diffprivlib.models.GaussianNB
        model. If model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: A name specifying the type of model, e.g., \
        naive_bayes
        :type model_name: `str`
        :param model_spec: A dictionary contains model specification
        :type model_spec: `dict`
        :param model: Compiled GaussianNB model
        :type model: `diffprivlib.models.GaussianNB`
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a Naive Bayes model.
        :type kwargs: `dict`
        """
        if model:
            if not issubclass(type(model), GaussianNB):
                raise ValueError(
                    "Compiled GaussianNB needs to be provided"
                    "(diffprivlib.models.GaussianNB). "
                    "Type provided" + str(type(model))
                )

        super().__init__(model_name, model_spec, sklearn_model=model, **kwargs)

        self.old_vals = {"theta": None, "var": None, "class_count": None}
        self.model_trained = False
        self.model_type = "Sklearn-NaiveBayes"

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data a tuple given in the form \
        (x_train, y_train).
        :type train_data: `np.ndarray`
        :param fit_params: Not used, present here for API consistency by \
        convention.
        :return: None
        """
        self.model.fit(train_data[0], train_data[1])
        self.model_trained = True

    def update_model(self, model_update, **kwargs):
        """
        Update GaussianNB model with provided model_update.

        :param model_update: `ModelUpdate` object that contains \
        the required information to update the GaussianNB model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if model_update.get("class_count") is None:
            return

        self.model.class_count_ = model_update.get("class_count")
        self.model.theta_ = model_update.get("theta")

        try:
            self.model.sigma_ = model_update.get("var")
        except AttributeError:
            self.model.var_ = model_update.get("var")

        self._store_old_values(model_update.get("theta"), model_update.get("var"), model_update.get("class_count"))

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        if (
            not self.model_trained
            or not hasattr(self.model, "class_count_")
            or np.all(self.old_vals["class_count"] == self.model.class_count_)
        ):
            return ModelUpdate(theta=None, var=None, class_count=None)

        try:
            model_var = self.model.var_
        except AttributeError:
            model_var = self.model.sigma_

        if self.old_vals["class_count"] is None:
            return ModelUpdate(theta=self.model.theta_, var=model_var, class_count=self.model.class_count_)

        class_count_update = self.model.class_count_ - self.old_vals["class_count"]

        nm = self.model.class_count_[:, np.newaxis]
        n = self.old_vals["class_count"][:, np.newaxis]
        m = class_count_update[:, np.newaxis]

        theta_update = self.model.theta_ * nm - self.old_vals["theta"] * n
        theta_update = theta_update / m

        var_update = (
            nm * model_var - n * self.old_vals["var"] - (n * m) / nm * (self.old_vals["theta"] - theta_update) ** 2
        )

        return ModelUpdate(theta=theta_update, var=var_update, class_count=class_count_update)

    def evaluate(self, test_dataset, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, test) or a datagenerator of of type `keras.utils.Sequence`, 
        `keras.preprocessing.image.ImageDataGenerator`
        :type test_dataset: `np.ndarray`

        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test)

        else:
            raise ModelException("Invalid test dataset!")

    def evaluate_model(self, x, y, **kwargs):
        """
        Evaluates the model given test data x and the corresponding labels y.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding true labels to x
        :type y: `np.ndarray`
        :param kwargs: Dictionary of model-specific arguments for evaluating \
        models. For example, sample weights accepted by model.score.
        :return: Dictionary with all evaluation metrics provided by \
        specific implementation.
        :rtype: `dict`
        """
        acc = {"score": self.model.score(x, y, **kwargs)}
        return acc

    def _store_old_values(self, theta, var, class_count):
        """
        Store old values of training parameters to allow reconstruction of
        parameters for newly-trained examples when model update is sought.
        """
        self.old_vals = {"theta": theta, "var": var, "class_count": class_count}

    @staticmethod
    def load_model_from_spec(model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains the following items: model_spec['model_definition']
        contains the model definition as type diffprivlib.GaussianNB.

        :param model_spec: Model specification contains \
        a complied sklearn model.
        :param model_spec: `dict`
        :return: model
        :rtype: `sklearn.cluster`
        """
        model = None
        try:
            if "model_definition" in model_spec:
                path = model_spec["model_definition"]
                model = SKLearnPersistence.load_model(path)
                
                if not issubclass(type(model), GaussianNB):
                    raise ValueError(
                        "Provided complied model in model_spec "
                        "should be of type GaussianNB."
                        "Instead they are:" + str(type(model))
                    )
        except Exception as ex:
            raise ModelInitializationException("Model specification was " "badly formed. ", str(ex))
        return model



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class PipelineVersionOutputEntity(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, content_href=None, parent=None):
        """
        PipelineVersionOutputEntity - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'content_href': 'str',
            'parent': 'PipelineVersionOutputEntityParent'
        }

        self.attribute_map = {
            'content_href': 'contentHref',
            'parent': 'parent'
        }

        self._content_href = content_href
        self._parent = parent

    @property
    def content_href(self):
        """
        Gets the content_href of this PipelineVersionOutputEntity.
        Href to the binary content of the pipeline. Make GET on it to download this file

        :return: The content_href of this PipelineVersionOutputEntity.
        :rtype: str
        """
        return self._content_href

    @content_href.setter
    def content_href(self, content_href):
        """
        Sets the content_href of this PipelineVersionOutputEntity.
        Href to the binary content of the pipeline. Make GET on it to download this file

        :param content_href: The content_href of this PipelineVersionOutputEntity.
        :type: str
        """

        self._content_href = content_href

    @property
    def parent(self):
        """
        Gets the parent of this PipelineVersionOutputEntity.


        :return: The parent of this PipelineVersionOutputEntity.
        :rtype: PipelineVersionOutputEntityParent
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """
        Sets the parent of this PipelineVersionOutputEntity.


        :param parent: The parent of this PipelineVersionOutputEntity.
        :type: PipelineVersionOutputEntityParent
        """

        self._parent = parent

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from abc import ABC, abstractmethod

import logging


logger = logging.getLogger(__name__)

class Serializer(ABC):
    """
    Abstract class for Serializer
    """

    @abstractmethod
    def serialize(self):
        """
        Serialize 

        :return: serialized byte stream
        :rtype: `b[]`
        """
        pass

    @abstractmethod
    def deserialize(self, serialization):
        """
        Deserialize 

        :param serialization: Serialization of the object
        :return: deserialized object
        """
        pass



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from typing import List, Dict

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.utils.autoai.errors import ContainerTypeNotSupported
from ibm_watson_machine_learning.helpers.connections import (DataConnection, ContainerLocation, S3Connection,
                                                             S3Location, FSLocation, AssetLocation)
from ibm_watson_machine_learning.utils.autoai.utils import is_ipython
from ibm_watson_machine_learning.foundation_models.utils import PromptTuningParams

import datetime
import numpy as np


class PromptTuner:
    id: str = None
    _client: APIClient = None
    _training_metadata: dict = None

    def __init__(self,
                 name: str,
                 task_id: str,
                 *,
                 description: str = None,
                 base_model: str = None,
                 accumulate_steps: int = None,
                 batch_size: int = None,
                 init_method: str = None,
                 init_text: str = None,
                 learning_rate: float = None,
                 max_input_tokens: int = None,
                 max_output_tokens: int = None,
                 num_epochs: int = None,
                 verbalizer: str = None,
                 tuning_type: str = None,
                 auto_update_model: bool = True,
                 group_by_name: bool = None):

        self.name = name
        self.description = description if description else "Prompt tuning with SDK"
        self.auto_update_model = auto_update_model
        self.group_by_name = group_by_name

        base_model = {'model_id': base_model}

        self.prompt_tuning_params = PromptTuningParams(base_model=base_model,
                                                       accumulate_steps=accumulate_steps,
                                                       batch_size=batch_size,
                                                       init_method=init_method,
                                                       init_text=init_text,
                                                       learning_rate=learning_rate,
                                                       max_input_tokens=max_input_tokens,
                                                       max_output_tokens=max_output_tokens,
                                                       num_epochs=num_epochs,
                                                       task_id=task_id,
                                                       tuning_type=tuning_type,
                                                       verbalizer=verbalizer)

        if not isinstance(self.name, str):
            raise WMLClientError(f"'name' param expected string, but got {type(self.name)}: {self.name}")

        if self.description and (not isinstance(self.description, str)):
            raise WMLClientError(f"'description' param expected string, but got {type(self.description)}: "
                                 f"{self.description}")

        if self.auto_update_model and (not isinstance(self.auto_update_model, bool)):
            raise WMLClientError(f"'auto_update_model' param expected bool, but got {type(self.auto_update_model)}: "
                                 f"{self.auto_update_model}")

        if self.group_by_name and (not isinstance(self.group_by_name, bool)):
            raise WMLClientError(f"'group_by_name' param expected bool, but got {type(self.group_by_name)}: "
                                 f"{self.group_by_name}")

    def run(self,
            training_data_references: List['DataConnection'],
            training_results_reference: DataConnection = None,
            background_mode=False
            ) -> dict:
        """Run a prompt tuning process of foundation model on top of the training data referenced by DataConnection.

        :param training_data_references: data storage connection details to inform where training data is stored
        :type training_data_references: list[DataConnection]

        :param training_results_reference: data storage connection details to store pipeline training results
        :type training_results_reference: DataConnection, optional

        :param background_mode: indicator if fit() method will run in background (async) or (sync)
        :type background_mode: bool, optional

        :return: run details
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment
            from ibm_watson_machine_learning.helpers import DataConnection, S3Location

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)

            prompt_tuner.run(
                training_data_connection=[DataConnection(
                    connection_asset_id=connection_id,
                    location=S3Location(
                        bucket='prompt_tuning_data',
                        path='pt_train_data.json')
                    )
                )]
                background_mode=False)
            """
        WMLResource._validate_type(training_data_references, "training_data_references", list, mandatory=True)
        WMLResource._validate_type(training_results_reference, "training_results_reference", object, mandatory=False)

        for source_data_connection in [training_data_references]:
            if source_data_connection:
                self._validate_source_data_connections(source_data_connection)

        training_results_reference = self._determine_result_reference(results_reference=training_results_reference,
                                                                      data_references=training_data_references)

        self._initialize_training_metadata(training_data_references, test_data_references=None,
                                               training_results_reference=training_results_reference)

        tuning_details = self._client.training.run(meta_props=self._training_metadata, asynchronous=background_mode)
        self.id = self._client.training.get_id(tuning_details)

        return self._client.training.get_details(self.id)  # TODO improve the background_mode = False option

    def _initialize_training_metadata(self,
                                          training_data_references: List['DataConnection'],
                                          test_data_references: List['DataConnection'] = None,
                                          training_results_reference: DataConnection = None,
                                          ):

        self._training_metadata = {
            self._client.training.ConfigurationMetaNames.TAGS: self._get_tags(),

            self._client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES:
                [connection._to_dict() for connection in training_data_references],

            self._client.training.ConfigurationMetaNames.NAME: f"{self.name[:100]}",
            self._client.training.ConfigurationMetaNames.PROMPT_TUNING: self.prompt_tuning_params.to_dict()
        }
        if test_data_references:
            self._training_metadata[self._client.training.ConfigurationMetaNames.TEST_DATA_REFERENCES] = [
                connection._to_dict() for connection in test_data_references]
        if training_results_reference:
            self._training_metadata[
                self._client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE] = \
                training_results_reference._to_dict()

        if self.description:
            self._training_metadata[
                self._client.training.ConfigurationMetaNames.DESCRIPTION] = f"{self.description}"

        if self.auto_update_model is not None:
            self._training_metadata[
                self._client.training.ConfigurationMetaNames.AUTO_UPDATE_MODEL] = self.auto_update_model

    def _validate_source_data_connections(self, source_data_connections):
        for data_connection in source_data_connections:
            if isinstance(data_connection.location, ContainerLocation):
                if self._client.ICP:
                    raise ContainerTypeNotSupported()  # block Container type on CPD
                elif isinstance(data_connection.connection, S3Connection):
                    # note: remove S3 inline credential from data asset before training
                    data_connection.connection = None
                    if hasattr(data_connection.location, 'bucket'):
                        delattr(data_connection.location, 'bucket')
                    # --- end note
            if isinstance(data_connection.connection, S3Connection) and isinstance(data_connection.location,
                                                                                   AssetLocation):
                # note: remove S3 inline credential from data asset before training
                data_connection.connection = None

                for s3_attr in ['bucket', 'path']:
                    if hasattr(data_connection.location, s3_attr):
                        delattr(data_connection.location, s3_attr)
                # --- end note

        return source_data_connections

    def _determine_result_reference(self, results_reference, data_references, result_path="default_tuning_output"):
        # note: if user did not provide results storage information, use default ones
        if results_reference is None:
            if self._client.ICP:
                location = FSLocation(path="/{option}/{id}/assets/wx_prompt_tune")
                if self._client.default_project_id is None:
                    location.path = location.path.format(option='spaces',
                                                         id=self._client.default_space_id)

                else:
                    location.path = location.path.format(option='projects',
                                                         id=self._client.default_project_id)
                results_reference = DataConnection(
                    connection=None,
                    location=location
                )

            else:
                if isinstance(data_references[0].location, S3Location):
                    results_reference = DataConnection(
                        connection=data_references[0].connection,
                        location=S3Location(bucket=data_references[0].location.bucket,
                                            path=".")
                    )

                elif isinstance(data_references[0].location, AssetLocation):
                    connection_id = data_references[0].location._get_connection_id(self._client)

                    if connection_id is not None:
                        results_reference = DataConnection(
                            connection_asset_id=connection_id,
                            location=S3Location(
                                bucket=data_references[0].location._get_bucket(self._client),
                                path=result_path)
                        )

                    else:  # set container output location when default DAta Asset is as a train ref
                        results_reference = DataConnection(
                            location=ContainerLocation(path=result_path))

                else:
                    results_reference = DataConnection(location=ContainerLocation(path=result_path))
        # -- end note

        # note: validate location types:
        if self._client.ICP:
            if not isinstance(results_reference.location,
                              FSLocation):
                raise TypeError('Unsupported results location type. Results reference can be stored on FSLocation.')
        else:
            if not isinstance(results_reference.location,
                              (S3Location, ContainerLocation)):
                raise TypeError('Unsupported results location type. Results reference can be stored'
                                ' only on S3Location or ContainerLocation.')
        # -- end note
        return results_reference

    def _get_tags(self):

        tags = ['prompt_tuning']
        if self.group_by_name is not None and self.group_by_name:

            for training in self._client.training.get_details(tag_value='prompt_tuning')['resources']:
                if training['metadata'].get('name') == self.name:
                    # Find recent tags related to 'name'
                    tags = list(set(tags) | set(training['metadata'].get('tags')))
                    break

            if tags != ['prompt_tuning']:
                self._client.generate_ux_tag = False

        return tags

    @staticmethod
    def _get_last_iteration_metrics_for_each_epoch(tuning_details):

        last_iteration_metrics_for_each_epoch = []
        for ind in range(len(tuning_details['entity']['status']['metrics'])):
            if ind == 0:
                last_iteration_metrics_for_each_epoch.append(tuning_details['entity']['status']['metrics'][0])
            else:
                if tuning_details['entity']['status']['metrics'][ind]['ml_metrics']['epoch'] == \
                        tuning_details['entity']['status']['metrics'][ind - 1]['ml_metrics']['epoch']:
                    last_iteration_metrics_for_each_epoch.pop()
                    last_iteration_metrics_for_each_epoch.append(tuning_details['entity']['status']['metrics'][ind])
                else:
                    last_iteration_metrics_for_each_epoch.append(tuning_details['entity']['status']['metrics'][ind])
        return last_iteration_metrics_for_each_epoch

    @staticmethod
    def _get_average_loss_score_for_each_epoch(tuning_details):

        scores = []
        temp_score = []
        epoch = 0
        if "data" in tuning_details['entity']['status']['metrics'][0]:
            for ind, metric in enumerate(tuning_details['entity']['status']['metrics']):
                if int(metric['data']['epoch']) == epoch:
                    temp_score.append(metric['data']['value'])
                else:
                    epoch += 1
                    scores.append(np.average(temp_score))
                    temp_score = [metric['data']['value']]
            scores.append(np.average(temp_score))
        else:
            for ind, metric in enumerate(tuning_details['entity']['status']['metrics']):
                if int(metric['ml_metrics']['epoch']) == epoch:
                    temp_score.append(metric['ml_metrics']['loss'])
                else:
                    epoch += 1
                    scores.append(np.average(temp_score))
                    temp_score = [metric['ml_metrics']['loss']]
            scores.append(np.average(temp_score))
        return scores

    @staticmethod
    def _get_first_and_last_iteration_metrics_for_each_epoch(tuning_details):

        first_and_last_iteration_metrics_for_each_epoch = []
        first_iteration = True

        tuning_metrics = tuning_details['entity']['status']['metrics']
        for ind in range(len(tuning_metrics)):
            if ind == 0:
                first_and_last_iteration_metrics_for_each_epoch.append(tuning_metrics[ind])
                first_and_last_iteration_metrics_for_each_epoch.append(tuning_metrics[ind])
                first_iteration = False
            elif first_iteration:
                first_and_last_iteration_metrics_for_each_epoch.append(tuning_metrics[ind])
                first_iteration = False
            else:
                if tuning_metrics[ind].get("data", tuning_metrics[ind].get("ml_metrics"))['epoch'] == tuning_metrics[ind - 1].get("data", tuning_metrics[ind-1].get("ml_metrics"))['epoch']:
                    first_and_last_iteration_metrics_for_each_epoch.pop()
                    first_and_last_iteration_metrics_for_each_epoch.append(tuning_metrics[ind])
                else:
                    first_and_last_iteration_metrics_for_each_epoch.append(tuning_metrics[ind])
                    first_iteration = True
        return first_and_last_iteration_metrics_for_each_epoch

    def get_params(self) -> dict:
        """Get configuration parameters of PromptTuner.

        :return: PromptTuner parameters
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)

            prompt_tuner.get_params()

            # Result:
            #
            # {'base_model': {'name': 'google/flan-t5-xl'},
            #  'task_id': 'summarization',
            #  'name': 'Prompt Tuning of Flan T5 model',
            #  'auto_update_model': False,
            #  'group_by_name': False}
        """

        params = self.prompt_tuning_params.to_dict()
        params['name'] = self.name
        params['description'] = self.description
        params['auto_update_model'] = self.auto_update_model
        params['group_by_name'] = self.group_by_name
        return params

    #####################
    #   Run operations  #
    #####################

    def get_run_status(self):
        """Check status/state of initialized Prompt Tuning run if ran in background mode.

        :return: Prompt tuning run status
        :rtype: str

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)
            prompt_tuner.run(...)

            prompt_tuner.get_run_details()

            # Result:
            # 'completed'
        """
        if self.id is None:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_not_scheduled"))

        return self._client.training.get_status(training_uid=self.id).get('state')

    def get_run_details(self, include_metrics: bool = False) -> dict:
        """Get prompt tuning run details.

        :param include_metrics: indicates to include metrics in the training details output
        :type include_metrics: bool, optional

        :return: Prompt tuning details
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)
            prompt_tuner.run(...)

            prompt_tuner.get_run_details()
        """
        if self.id is None:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_not_scheduled"))

        details = self._client.training.get_details(training_uid=self.id)

        if include_metrics:
            try:
                details["entity"]["status"]["metrics"] = self._get_metrics_data_from_property_or_file(details)
            except KeyError:
                pass
            finally:
                return details

        if details['entity']['status'].get('metrics', False):
            del details['entity']['status']['metrics']

        return details

    def _get_metrics_data_from_property_or_file(self, details: Dict) -> Dict:
        path = details["entity"]["status"]["metrics"][0]["context"]["prompt_tuning"]["metrics_location"]
        results_reference = details["entity"]['results_reference']
        conn = DataConnection._from_dict(results_reference)
        conn._wml_client = self._client
        metrics_data = conn._download_json_file(path)

        return metrics_data

    def plot_learning_curve(self):
        """Plot learning curves.

        .. note ::
            Available only for Jupyter notebooks.

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)
            prompt_tuner.run(...)

            prompt_tuner.plot_learning_curve()
        """
        if not is_ipython():
            raise WMLClientError("Function `plot_learning_curve` is available only for Jupyter notebooks.")
        from ibm_watson_machine_learning.utils.autoai.incremental import plot_learning_curve
        import matplotlib.pyplot as plt
        
        tuning_details = self.get_run_details(include_metrics=True)

        if 'metrics' in tuning_details['entity']['status']:

            # average loss score for each epoch
            scores = self._get_average_loss_score_for_each_epoch(tuning_details=tuning_details)

            # date_time from the first and last iteration on each epoch
            if "data" in tuning_details['entity']['status']['metrics'][0]:
                date_times = [datetime.datetime.strptime(m_obj["data"]['timestamp'], '%Y-%m-%dT%H:%M:%S.%f')
                              for m_obj in
                              self._get_first_and_last_iteration_metrics_for_each_epoch(tuning_details=tuning_details)]
            else:
                date_times = [datetime.datetime.strptime(m_obj['timestamp'], '%Y-%m-%dT%H:%M:%S.%f%z')
                              for m_obj in self._get_first_and_last_iteration_metrics_for_each_epoch(tuning_details=tuning_details)]

            elapsed_time = []
            for i in range(1, len(date_times), 2):
                elapsed_time.append((date_times[i] - date_times[i - 1]).total_seconds())

            fig, axes = plt.subplots(1, 3, figsize=(18, 4))
            if scores:
                plot_learning_curve(fig=fig, axes=axes, scores=scores, fit_times=elapsed_time,
                                    xlabels={'first_xlabel': 'Epochs', 'second_xlabel': 'Epochs'},
                                    titles={'first_plot': 'Loss function'})
        else:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_no_metrics"))

    def summary(self, scoring: str = 'loss') -> 'DataFrame':
        """Print PromptTuner models details (prompt-tuned models).

        :param scoring: scoring metric which user wants to use to sort pipelines by,
            when not provided use loss one
        :type scoring: string, optional

        :return: computed models and metrics
        :rtype: pandas.DataFrame

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)
            prompt_tuner.run(...)

            prompt_tuner.summary()

            # Result:
            #                          Enhancements            Base model  ...         loss
            #       Model Name
            # Prompt_tuned_M_1      [prompt_tuning]     google/flan-t5-xl  ...     0.449197
        """

        if self.id is None:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_not_scheduled"))

        from pandas import DataFrame

        details = self.get_run_details(include_metrics=True)

        metrics = details['entity']['status'].get('metrics', [{}])[0]
        is_ml_metrics = 'data' in metrics or 'ml_metrics' in metrics

        if not is_ml_metrics:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_no_metrics"))

        columns = ['Model Name', 'Enhancements', 'Base model', 'Auto store', 'Epochs', scoring]
        values = []
        model_name = 'model_' + self.id
        base_model_name = None
        epochs = None
        enhancements = []
        if scoring == 'loss':
            model_metrics = [self._get_average_loss_score_for_each_epoch(tuning_details=details)[-1]]
        else:
            if "data" in details['entity']['status']['metrics'][0]:
                model_metrics = [details['entity']['status'].get('metrics', [{}])[-1].get('data', {})[scoring]]
            else:
                model_metrics = [details['entity']['status'].get('metrics', [{}])[-1].get('ml_metrics', {})[scoring]]

        if 'prompt_tuning' in details['entity']:
            enhancements = [details['entity']['prompt_tuning']['tuning_type']]
            base_model_name = details['entity']['prompt_tuning']['base_model']['model_id']
            epochs = details['entity']['prompt_tuning']['num_epochs']

        values.append((
                [model_name] +
                [enhancements] +
                [base_model_name] +
                [details['entity']['auto_update_model']] +
                [epochs] +
                model_metrics
             ))

        summary = DataFrame(data=values, columns=columns)
        summary.set_index('Model Name', inplace=True)

        return summary

    def get_model_id(self):
        """Get model id.

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)
            prompt_tuner.run(...)

            prompt_tuner.get_model_id()
        """

        run_details = self.get_run_details()
        if run_details['entity']['auto_update_model']:
            return run_details['entity']['model_id']
        else:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_no_model_id"))

    def cancel_run(self, hard_delete=False) -> None:
        """Cancels or deletes a Prompt Tuning run.

        :param hard_delete: When True then the completed or cancelled prompt tuning run is deleted,
                            if False then the current run is canceled. Default: False
        :type hard_delete: bool, optional
        """
        if self.id is None:
            raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_not_scheduled"))

        self._client.training.cancel(training_uid=self.id, hard_delete=hard_delete)

    def get_data_connections(self) -> List['DataConnection']:
        """Create DataConnection objects for further user usage
            (eg. to handle data storage connection).

        :return: list of DataConnections
        :rtype: list['DataConnection']

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment
            experiment = TuneExperiment(credentials, ...)
            prompt_tuner = experiment.prompt_tuner(...)
            prompt_tuner.run(...)

            data_connections = prompt_tuner.get_data_connections()
        """

        training_data_references = self.get_run_details()['entity']['training_data_references']

        data_connections = [
            DataConnection._from_dict(_dict=data_connection) for data_connection in training_data_references]

        for data_connection in data_connections:
            data_connection.set_client(self._client)
            data_connection._run_id = self.id

        return data_connections



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class EvaluationDefinitionExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, method=None, metrics=None):
        """
        EvaluationDefinitionExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'method': 'str',
            'metrics': 'list[MetricObjectExperiments]'
        }

        self.attribute_map = {
            'method': 'method',
            'metrics': 'metrics'
        }

        self._method = method
        self._metrics = metrics

    @property
    def method(self):
        """
        Gets the method of this EvaluationDefinitionExperiments.


        :return: The method of this EvaluationDefinitionExperiments.
        :rtype: str
        """
        return self._method

    @method.setter
    def method(self, method):
        """
        Sets the method of this EvaluationDefinitionExperiments.


        :param method: The method of this EvaluationDefinitionExperiments.
        :type: str
        """
        allowed_values = ["binary", "regression", "multiclass"]
        if method not in allowed_values:
            raise ValueError(
                "Invalid value for `method` ({0}), must be one of {1}"
                .format(method, allowed_values)
            )

        self._method = method

    @property
    def metrics(self):
        """
        Gets the metrics of this EvaluationDefinitionExperiments.


        :return: The metrics of this EvaluationDefinitionExperiments.
        :rtype: list[MetricObjectExperiments]
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """
        Sets the metrics of this EvaluationDefinitionExperiments.


        :param metrics: The metrics of this EvaluationDefinitionExperiments.
        :type: list[MetricObjectExperiments]
        """

        self._metrics = metrics

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging

import joblib
import time

from ibmfl.exceptions import FLException
logger = logging.getLogger(__name__)


class SKLearnPersistence():

    @staticmethod
    def load_model(full_path):
        try:
           model = joblib.load(full_path)
           return model
        except Exception as e:
           logger.exception(e)
           raise FLException("Unable to load model.")

    @staticmethod
    def model_filename(file):
        return "{}_{}.pickle".format(file, time.time())
    
    @staticmethod
    def save_model(model, full_path):
        with open(full_path, "wb") as f:
            joblib.dump(model, f)
        logger.info("Model saved in path: %s.", full_path)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

__all__ = [
    "TuneRuns"
]

from pandas import DataFrame

from ibm_watson_machine_learning.foundation_models.prompt_tuner import PromptTuner
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning import APIClient


class TuneRuns:
    """TuneRuns class is used to work with historical PromptTuner runs.

    :param client: APIClient to handle service operations
    :type client: APIClient

    :param filter: filter, user can choose which runs to fetch specifying tuning name
    :type filter: str, optional

    :param limit: int number of records to be returned
    :type limit: int
    """

    def __init__(self, client: 'APIClient', filter: str = None, limit: int = 50) -> None:

        self.client = client
        self.tuning_name = filter
        self.limit = limit

    def __call__(self, *, filter: str = None, limit: int = 50) -> 'TuneRuns':
        self.tuning_name = filter
        self.limit = limit
        return self

    def list(self) -> 'DataFrame':
        """Lists historical runs with status. If user has a lot of runs stored in the service,
        it may take long time to fetch all the information. If there is no limit set,
        get last 50 records.

        :return: Pandas DataFrame with runs IDs and state
        :rtype: pandas.DataFrame

        **Examples**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(...)
            df = experiment.runs.list()
        """

        runs_details = self.client.training.get_details(get_all=True if self.tuning_name else False,
                                                        limit=None if self.tuning_name else self.limit,
                                                        training_type='prompt_tuning',
                                                        _internal=True)

        columns = ['timestamp', 'run_id', 'state', 'prompt tuning name']

        records = []
        for run in runs_details['resources']:
            if len(records) >= self.limit:
                break

            if {'entity', 'metadata'}.issubset(run.keys()):

                timestamp = run['metadata'].get('modified_at')
                run_id = run['metadata'].get('id', run['metadata'].get('guid'))
                state = run['entity'].get('status', {}).get('state')
                tuning_name = run['entity'].get('name', 'Unknown')

                record = [timestamp, run_id, state, tuning_name]

                if self.tuning_name is None or (self.tuning_name and self.tuning_name == tuning_name):
                    records.append(record)

        runs = DataFrame(data=records, columns=columns)
        return runs.sort_values(by=["timestamp"], ascending=False)

    def get_tuner(self, run_id: str) -> PromptTuner:
        """Create instance of PromptTuner based on tuning run with specific run_id.

        :param run_id: ID of the run
        :type run_id: str

        :return: prompt tuner object
        :rtype: PromptTuner class instance

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(credentials, ...)
            historical_tuner = experiment.runs.get_tuner(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        """
        # note: normal scenario

        if not isinstance(run_id, str):
            raise WMLClientError(f"Provided run_id type was {type(run_id)} (should be a string)")

        entity = self.client.training.get_details(run_id).get('entity')
        if not entity:
            raise WMLClientError("Provided run_id was invalid")

        tuning_params = entity['prompt_tuning']

        prompt_tuner = PromptTuner(name=entity.get('name'),
                                   task_id=tuning_params.get('task_id'),
                                   description=entity.get('description'),
                                   base_model=tuning_params.get('base_model', {}).get('name'),
                                   accumulate_steps=tuning_params.get('accumulate_steps'),
                                   batch_size=tuning_params.get('batch_size'),
                                   init_method=tuning_params.get('init_method'),
                                   init_text=tuning_params.get('init_text'),
                                   learning_rate=tuning_params.get('learning_rate'),
                                   max_input_tokens=tuning_params.get('max_input_tokens'),
                                   max_output_tokens=tuning_params.get('max_output_tokens'),
                                   num_epochs=tuning_params.get('num_epochs'),
                                   tuning_type=tuning_params.get('tuning_type'),
                                   verbalizer=tuning_params.get('verbalizer'),
                                   auto_update_model=entity.get('auto_update_model'))

        prompt_tuner.id = run_id
        prompt_tuner._client = self.client
        return prompt_tuner

    def get_run_details(self, run_id: str = None, include_metrics: bool = False) -> dict:
        """Get run details. If run_id is not supplied, last run will be taken.

        :param run_id: ID of the run
        :type run_id: str, optional

        :param include_metrics: indicates to include metrics in the training details output
        :type include_metrics: bool, optional

        :return: run configuration parameters
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment
            experiment = TuneExperiment(credentials, ...)

            experiment.runs.get_run_details(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
            experiment.runs.get_run_details()
        """
        if run_id is None:
            details = self.client.training.get_details(limit=1, training_type='prompt_tuning', _internal=True).get('resources')[0]
        else:
            details = self.client.training.get_details(training_uid=run_id, _internal=True)

        if include_metrics:
            return details

        if details['entity']['status'].get('metrics', False):
            del details['entity']['status']['metrics']

        return details
